#include <vector>
#include <SDL.h>
#include <SDL_kanji.h>
#include "wx/wx.h"
#include "ist/bitmap.h"
using namespace std;


class F1BApp : public wxApp
{
public:
	int OnExit();
	bool OnInit();
};

DECLARE_APP(F1BApp)



class F1BFrame : public wxFrame
{
public:
	F1BFrame();
	void SelectBDF1(wxCommandEvent&);
	void SelectBDF2(wxCommandEvent&);
	void ExportF1B(wxCommandEvent&);

private:
	wxTextCtrl *text_bdf1;
	wxTextCtrl *text_bdf2;
	wxTextCtrl *text_font_size;

	DECLARE_EVENT_TABLE()
};







enum {
	TEXT_BDF1 = 100,
	TEXT_BDF2,
	BUTTON_BDF1,
	BUTTON_BDF2,
	BUTTON_EXPORT,
};

BEGIN_EVENT_TABLE( F1BFrame, wxFrame )
	EVT_BUTTON( BUTTON_BDF1, F1BFrame::SelectBDF1 )
	EVT_BUTTON( BUTTON_BDF2, F1BFrame::SelectBDF2 )
	EVT_BUTTON( BUTTON_EXPORT, F1BFrame::ExportF1B )
END_EVENT_TABLE()


F1BFrame::F1BFrame() :
	wxFrame(0, -1, wxT("bdf2f1b"), wxDefaultPosition, wxDefaultSize,
		wxSYSTEM_MENU | wxCAPTION | wxCLOSE_BOX)
{
	CreateStatusBar();

	wxPanel *panel = new wxPanel( this, -1, wxDefaultPosition, wxDefaultSize );
	wxBoxSizer *top_sizer = new wxBoxSizer(wxVERTICAL);

	{
		wxBoxSizer *sizer1 = new wxBoxSizer(wxHORIZONTAL);
		text_bdf1 = new wxTextCtrl( panel, TEXT_BDF1, wxT(""), wxDefaultPosition, wxSize(220, -1), wxTE_READONLY );
		wxButton *button_bdf1 = new wxButton( panel, BUTTON_BDF1, wxT("select bdf 1") );
		sizer1->Add( button_bdf1, 0, 0);
		sizer1->Add( 5, 0 );
		sizer1->Add( text_bdf1, 1, wxEXPAND);

		wxBoxSizer *sizer2 = new wxBoxSizer(wxHORIZONTAL);
		text_bdf2 = new wxTextCtrl( panel, TEXT_BDF2, wxT(""), wxDefaultPosition, wxSize(220, -1), wxTE_READONLY );
		wxButton *button_bdf2 = new wxButton( panel, BUTTON_BDF2, wxT("select bdf 2") );
		sizer2->Add( button_bdf2, 0, 0);
		sizer2->Add( 5, 0 );
		sizer2->Add( text_bdf2, 1, wxEXPAND);

		wxBoxSizer *sizer3 = new wxBoxSizer(wxHORIZONTAL);
		wxStaticText *text = new wxStaticText( panel, 0, wxT("font size") );
		text_font_size = new wxTextCtrl( panel, TEXT_BDF2, wxT("0"), wxDefaultPosition, wxSize(40, -1) );
		sizer3->Add( text, 0, wxALIGN_CENTER_VERTICAL );
		sizer3->Add( 5, 0 );
		sizer3->Add( text_font_size, 1, wxEXPAND);

		wxButton *button_export = new wxButton( panel, BUTTON_EXPORT, wxT("export") );

		top_sizer->Add( sizer1, 0, wxALL, 5 );
		top_sizer->Add( sizer2, 0, wxALL, 5 );
		top_sizer->Add( sizer3, 0, wxALL, 5 );
		top_sizer->Add( button_export, 0, wxALL, 5 );
	}

	panel->SetAutoLayout( TRUE );
	panel->SetSizer( top_sizer );
	panel->Fit();
	SetClientSize(panel->GetSize());
}

void F1BFrame::SelectBDF1(wxCommandEvent&)
{
	wxFileDialog file_dlg(this, wxT("select bdf 1"), wxT(""), wxT(""), wxT("*.bdf") );
	if(file_dlg.ShowModal() == wxID_OK)
		text_bdf1->SetValue( file_dlg.GetPath() );
}

void F1BFrame::SelectBDF2(wxCommandEvent&)
{
	wxFileDialog file_dlg(this, wxT("select bdf 2"), wxT(""), wxT(""), wxT("*.bdf") );
	if(file_dlg.ShowModal() == wxID_OK)
		text_bdf2->SetValue( file_dlg.GetPath() );
}








static void RLEWriteSameData(vector<unsigned char>& data, FILE *&out)
{
	if(data.empty())
		return;

	unsigned char length = data.size()+0x80;
	fwrite(&length, sizeof(unsigned char), 1, out);
	fwrite(&data.front(), sizeof(unsigned char), 1, out);

	data.clear();
}

static void RLEWriteDiffData(vector<unsigned char>& data, FILE *&out)
{
	if(data.empty())
		return;

	unsigned char length = data.size()-1;
	fwrite(&length, sizeof(unsigned char), 1, out);

	for(vector<unsigned char>::iterator p=data.begin(); p!=data.end(); ++p) {
		fwrite(&(*p), sizeof(unsigned char), 1, out);
	}

	data.clear();
}


static bool SaveF1B(ist::Bitmap& bmp, int font_size, const char *filename)
{
	if(bmp.empty())
		return false;

	int n = 0;
	while(n<font_size) { n+=8; }
	int char_bit = n*font_size;


	vector<unsigned char> data;
	int cell_x = bmp.getWidth()/font_size;
	int cell_y = bmp.getHeight()/font_size;

	ist::bRGBA black(0, 0, 0, 255);
	int count = 0;
	unsigned char fdata = 0;

	for(int i=0; i<cell_y; ++i ) {
		for(int j=0; j<cell_x; ++j ) {

			for(int k=font_size-1; k>=0; --k) {
				for(int l=0; l<font_size; ++l) {
					if(bmp[i*font_size+k][j*font_size+l]==black)
						fdata+=1<<(7-count);

					++count;
					if(count==8 || l==font_size-1) {
						data.push_back(fdata);
						fdata = 0;
						count = 0;
					}
				}
			}

		}
	}


	FILE *out = fopen(filename, "wb");
	if(out==0)
		return false;

	fwrite(&font_size, sizeof(int), 1, out);

	vector<unsigned char> same, diff;
	for(vector<unsigned char>::iterator p=data.begin(); p!=data.end(); ++p) {
		vector<unsigned char>::iterator next = (p+1);
		if(next!=data.end() && *p==*next && same.size()<0x7f) {
			same.push_back(*p);
			RLEWriteDiffData(diff, out);
		}
		else {
			if(!same.empty()) {
				RLEWriteSameData(same, out);
			}
			else {
				diff.push_back(*p);
				if(diff.size()==0x7f || next==data.end())
					RLEWriteDiffData(diff, out);
			}
		}
	}

	fclose(out);
	return true;
}






static SDL_Surface* CreateCharSurface(Kanji_Font *font, int index)
{
	if (font->moji[index]==0)
		return 0;

	SDL_Surface *cs = SDL_AllocSurface(SDL_SWSURFACE, font->k_size, font->k_size, 32, 0, 0, 0, 0);
	Uint32 bgcol = SDL_MapRGB(cs->format, 255, 255, 255);
	Uint32 fgcol = SDL_MapRGB(cs->format, 0, 0, 0);
	SDL_FillRect(cs, 0, bgcol);

	for (int y=0; y<font->k_size; ++y) {
		for (int x=0; x<font->k_size; ++x) {
			if (font->moji[index][y] & (1 << (font->k_size-x-1)))
				((Uint32*)cs->pixels)[(y*font->k_size)+x] = fgcol;
		}
	}

	return cs;
}

static void CopyCharSurface(ist::Bitmap &bmp, Kanji_Font *font, int font_size)
{
	for(int index=0; index<96*96+256; ++index) {
		SDL_Surface *cs = CreateCharSurface(font, index);
		if(cs==0)
			continue;

		int cx = index%64;
		int cy = index/64;
		SDL_PixelFormat *fmt = cs->format;

		for(int i=cy*font_size; i<(cy+1)*font_size; ++i ) {
			for(int j=cx*font_size; j<(cx+1)*font_size; ++j ) {
				Uint32 temp;
				Uint32 pixel = ((Uint32*)cs->pixels)[cs->w*(i-cy*font_size)+(j-cx*font_size)];

				temp = pixel&fmt->Rmask;
				temp = temp>>fmt->Rshift;
				temp = temp<<fmt->Rloss;
				bmp[i][j].r = (Uint8)temp;

				temp = pixel&fmt->Gmask;
				temp = temp>>fmt->Gshift;
				temp = temp<<fmt->Gloss;
				bmp[i][j].g = (Uint8)temp;

				temp = pixel&fmt->Bmask;
				temp = temp>>fmt->Bshift;
				temp = temp<<fmt->Bloss;
				bmp[i][j].b = (Uint8)temp;

				if(bmp[i][j].r==0)
					bmp[i][j].a = 255;
				else
					bmp[i][j].a = 0;
			}
		}

		SDL_FreeSurface(cs);
	}
}

void F1BFrame::ExportF1B(wxCommandEvent&)
{
	wxString& path1 = text_bdf1->GetValue();
	wxString& path2 = text_bdf2->GetValue();

	unsigned long font_size = 0;
	if(!text_font_size->GetValue().ToULong(&font_size) || font_size==0) {
		SetStatusText( wxT("invalid font size") );
		return;
	}

	Kanji_Font *font = Kanji_OpenFont(path1.c_str(), font_size);
	if(font==0) {
		SetStatusText( wxT("failed to open font") );
		return;
	}
	Kanji_AddFont(font, path2.c_str());


	wxFileDialog file_dlg(this, wxT("export f1b"), wxT(""), wxT(""), wxT("*.f1b"), wxSAVE );
	if(file_dlg.ShowModal() == wxID_OK) {
		SetStatusText( wxT("exporting f1b...") );

		ist::Bitmap bmp;
		bmp.resize(font_size*64, font_size*148);
		CopyCharSurface(bmp, font, font_size);
		SaveF1B(bmp, font_size, file_dlg.GetPath().c_str());

		SetStatusText( wxT("exporting tga...") );
		bmp.save( (file_dlg.GetPath()+wxT(".tga")).c_str() );

		SetStatusText( wxT("done") );
	}

	Kanji_CloseFont(font);
}







IMPLEMENT_APP(F1BApp)

int F1BApp::OnExit()
{
	SDL_Quit();
	return 0;
}

bool F1BApp::OnInit()
{
	SDL_Init(SDL_INIT_VIDEO);

	F1BFrame *frame = new F1BFrame();
	frame->Show(TRUE);
	SetTopWindow(frame);

	return true;
}



