#include "bitmap.h"

using namespace std;

namespace ist {

typedef unsigned char u_char;

typedef unsigned int u_int;

bool Bitmap::load(const char *filename)
{
	u_int length=strlen(filename);
	*this=Bitmap();

	if(    strncmp(&filename[length-3], "tga", 3)==0 && loadTGA(filename)
		|| strncmp(&filename[length-3], "bmp", 3)==0 && loadBMP(filename) )
		return true;
	else
		return false;
}

bool Bitmap::save(const char *filename) const
{
	u_int length=strlen(filename);

	if(    strncmp(&filename[length-3], "tga", 3)==0 && saveTGA(filename)
		|| strncmp(&filename[length-3], "bmp", 3)==0 && saveBMP(filename) )
		return true;
	else
		return false;
}




struct BMPHEAD {
	char B;
	char M;
	int file_size;
	short reserve1;
	short reserve2;
	int offset;

	BMPHEAD() {
		memset(this, 0, sizeof(*this));
		B='B';
		M='M';
		offset=54;
	}
};

struct BMPINFOHEAD {
	int header_size;
	int width;
	int height;
	short plane;
	short bits;
	int compression;
	int comp_image_size;
	int x_resolution;
	int y_resolution;
	int pallete_num;
	int important_pallete_num;

	BMPINFOHEAD() {
		memset(this, 0, sizeof(*this));
		header_size=sizeof(*this);
		plane=1;
		bits=24;
	}
};

struct TGAHEAD {
	u_char No_ID;
	u_char CMap_Type;
	u_char image_type;
	u_char CMap_Spec[5];
	short Ox;
	short Oy;
	short width;
	short height;
	u_char pixel;
	u_char IDesc_Type;

	TGAHEAD() {
		memset(this, 0, sizeof(*this));
		pixel=32;
		IDesc_Type=8;
	}
};


static bRGBA Read1Pixel(FILE *&read)
{
	bRGBA t;
	fread(&t.b, sizeof(u_char), 1, read);
	fread(&t.g, sizeof(u_char), 1, read);
	fread(&t.r, sizeof(u_char), 1, read);
	fread(&t.a, sizeof(u_char), 1, read);
	return t;
}


namespace RLECompress {
	vector<u_char> comp_pixel;


	static void WriteSameData(vector<bRGBA> &temp_pixel)
	{
		comp_pixel.push_back( temp_pixel.size()+0x80 );

		comp_pixel.push_back( temp_pixel[0].b );
		comp_pixel.push_back( temp_pixel[0].g );
		comp_pixel.push_back( temp_pixel[0].r );
		comp_pixel.push_back( temp_pixel[0].a );

		temp_pixel.clear();
	}

	static void WriteDifferentData(vector<bRGBA> &temp_pixel)
	{
		comp_pixel.push_back( temp_pixel.size()-1 );

		for(u_int d=0; d<temp_pixel.size(); d++) {
			comp_pixel.push_back( temp_pixel[d].b );
			comp_pixel.push_back( temp_pixel[d].g );
			comp_pixel.push_back( temp_pixel[d].r );
			comp_pixel.push_back( temp_pixel[d].a );
		}

		temp_pixel.clear();
	}

	static void Compress(const bRGBA *start, int width)
	{
		vector<bRGBA> same, diff;

		for(int i=0; i!=width; ++i, ++start) {
			const bRGBA *ip=start; ++ip;
			bRGBA dist=*start;

			if( i+1!=width && dist==*ip && same.size()<0x79 ) {
				same.push_back(dist);

				if(diff.size()!=0)
					WriteDifferentData(diff);
			}
			else  {
				if(same.size()>0x00) {
					WriteSameData(same);
				}
				else {
					diff.push_back(dist);

					if(diff.size()==0x79 )
						WriteDifferentData(diff);
				}
			}
		}

		if(same.size()!=0x00)
			WriteSameData(same);
		else if(diff.size()!=0)
			WriteDifferentData(diff);
	}
}



// BMP

bool Bitmap::loadBMP(const char *filename)
{
	u_int j;
	FILE *read;
	BMPHEAD head;
	BMPINFOHEAD infohead;

	read=fopen(filename, "rb");
	if (read==0) return false;

	fread(&head.B,         sizeof(char), 1, read);
	fread(&head.M,         sizeof(char), 1, read);
	fread(&head.file_size, sizeof(int) , 1, read);
	fread(&head.reserve1,  sizeof(short),1, read);
	fread(&head.reserve2,  sizeof(short),1, read);
	fread(&head.offset,    sizeof(int),  1, read);

	fread(&infohead.header_size,sizeof(int), 1, read);
	fread(&infohead.width,  sizeof(int), 1, read);
	fread(&infohead.height, sizeof(int), 1, read);
	fread(&infohead.plane,  sizeof(short), 1, read);
	fread(&infohead.bits,   sizeof(short), 1, read);
	fread(&infohead.compression,    sizeof(int), 1, read);
	fread(&infohead.comp_image_size,sizeof(int), 1, read);
	fread(&infohead.x_resolution,   sizeof(int), 1, read);
	fread(&infohead.y_resolution,   sizeof(int), 1, read);
	fread(&infohead.pallete_num,    sizeof(int), 1, read);
	fread(&infohead.important_pallete_num, sizeof(int), 1, read);

	if(head.B!='B' || head.M!='M' || infohead.bits!=24 ) return false;

	pixel.width=infohead.width;
	pixel.height=infohead.height;
	pixel.data.resize(pixel.width*pixel.height);

	for(int i=pixel.height-1; i>=0; --i) {
		for(int j=0; j<pixel.width; ++j) {
			fread(&pixel[i][j].b, sizeof(u_char), 1, read);
			fread(&pixel[i][j].g, sizeof(u_char), 1, read);
			fread(&pixel[i][j].r, sizeof(u_char), 1, read);

			pixel[i][j].a=pixel[i][j].r;
		//	pixel[i][j].a=255;
		}
	}
	fclose(read);

	return true;
}


bool Bitmap::saveBMP(const char *filename) const
{
	if(pixel.data.empty()) return false;

	FILE *write=fopen(filename, "wb");
	BMPHEAD head;
	BMPINFOHEAD infohead;

	head.file_size=sizeof(BMPHEAD)+sizeof(BMPINFOHEAD)+pixel.width*pixel.height*3;
	infohead.width=pixel.width;
	infohead.height=pixel.height;

	fwrite(&head.B,         sizeof(char), 1, write);
	fwrite(&head.M,         sizeof(char), 1, write);
	fwrite(&head.file_size, sizeof(int) , 1, write);
	fwrite(&head.reserve1,  sizeof(short),1, write);
	fwrite(&head.reserve2,  sizeof(short),1, write);
	fwrite(&head.offset,    sizeof(int),  1, write);

	fwrite(&infohead.header_size,sizeof(int), 1, write);
	fwrite(&infohead.width,  sizeof(int), 1, write);
	fwrite(&infohead.height, sizeof(int), 1, write);
	fwrite(&infohead.plane,  sizeof(short), 1, write);
	fwrite(&infohead.bits,   sizeof(short), 1, write);
	fwrite(&infohead.compression,    sizeof(int), 1, write);
	fwrite(&infohead.comp_image_size,sizeof(int), 1, write);
	fwrite(&infohead.x_resolution,   sizeof(int), 1, write);
	fwrite(&infohead.y_resolution,   sizeof(int), 1, write);
	fwrite(&infohead.pallete_num,    sizeof(int), 1, write);
	fwrite(&infohead.important_pallete_num, sizeof(int), 1, write);

	for(int i=pixel.height-1; i>=0; --i) {
		for(int j=0; j<pixel.width; ++j) {
			fwrite(&pixel[i][j].b, sizeof(u_char), 1, write);
			fwrite(&pixel[i][j].g, sizeof(u_char), 1, write);
			fwrite(&pixel[i][j].r, sizeof(u_char), 1, write);
		}
	}
	fclose(write);

	return true;
}





// TGA

bool Bitmap::loadTGA(const char *filename)
{
	FILE *read;
	TGAHEAD head;

	read=fopen(filename, "rb");
	if (read==0) return false;

	fread(&head.No_ID,     sizeof(u_char), 1, read);
	fread(&head.CMap_Type, sizeof(u_char), 1, read);
	fread(&head.image_type,sizeof(u_char), 1, read);
	fread(&head.CMap_Spec, sizeof(u_char), 5, read);
	fread(&head.Ox,        sizeof(short), 1, read);
	fread(&head.Oy,        sizeof(short), 1, read);
	fread(&head.width,     sizeof(short), 1, read);
	fread(&head.height,    sizeof(short), 1, read);
	fread(&head.pixel,     sizeof(u_char), 1, read);
	fread(&head.IDesc_Type,sizeof(u_char), 1, read);

	if(head.pixel!=32) return false;

	pixel.width=head.width;
	pixel.height=head.height;
	pixel.data.resize(pixel.width*pixel.height);

	for(int i=pixel.height-1; i>=0; --i) {
		if(head.image_type==2) {
			for(int j=0; j<pixel.width; j++)
				pixel[i][j]=Read1Pixel(read);
		}
		else if(head.image_type==10) {
			int loaded=0;

			while( loaded < pixel.width ) {
				u_char dist;

				fread(&dist, sizeof(u_char), 1, read);

				if( dist<0x80) {
					for(int j=0; j<dist+1; ++j, ++loaded)
						pixel[i][loaded]=Read1Pixel(read);
				}
				else {
					bRGBA t=Read1Pixel(read);

					for(int j=0x80; j<dist+1; ++j, ++loaded)
						pixel[i][loaded]=t;
				}
			}
		}
	}
	fclose(read);

	return true;
}


bool Bitmap::saveTGA(const char *filename) const
{
	if(pixel.data.empty()) return false;

	FILE *write;
	TGAHEAD head;

	head.width=pixel.width;
	head.height=pixel.height;
	head.image_type=10;

	write=fopen(filename, "wb");
	fwrite(&head.No_ID,     sizeof(u_char), 1, write);
	fwrite(&head.CMap_Type, sizeof(u_char), 1, write);
	fwrite(&head.image_type,sizeof(u_char), 1, write);
	fwrite(&head.CMap_Spec, sizeof(u_char), 5, write);
	fwrite(&head.Ox,        sizeof(short), 1, write);
	fwrite(&head.Oy,        sizeof(short), 1, write);
	fwrite(&head.width,     sizeof(short), 1, write);
	fwrite(&head.height,    sizeof(short), 1, write);
	fwrite(&head.pixel,     sizeof(u_char), 1, write);
	fwrite(&head.IDesc_Type,sizeof(u_char), 1, write);

	for(int i=pixel.height-1; i>=0; --i)
		RLECompress::Compress(pixel[i], pixel.width);

	for(vector<u_char>::iterator p=RLECompress::comp_pixel.begin(); p!=RLECompress::comp_pixel.end(); ++p)
			fwrite(&(*p), sizeof(u_char), 1, write);

	fclose(write);

	vector<u_char>().swap(RLECompress::comp_pixel);

	return true;
}


}	// ist


